@@ -1,11 +1,9 @@
-### pseudoinverse.R  (2004-01-15)
+### pseudoinverse.R  (2004-09-15)
 ###
 ###    Computation of the Pseudoinverse of a Matrix
 ###
 ### Copyright 2003-04 Korbinian Strimmer
 ###
-### This is essentially a proxy to "ginv" from the MASS package
-### by B. Venables and B. Ripley.
 ###
 ### This file is part of the `GeneTS' library for R and related languages.
 ### It is made available under the terms of the GNU General Public
@@ -26,7 +24,32 @@
 
 
 # pseudoinverse of a matrix
-pseudoinverse <- function (m, tol = sqrt(.Machine$double.eps))
+pseudoinverse <- function (m, tol)
 {
-    return( ginv(m, tol) )
+    if (length(dim(m)) > 2 || !(is.numeric(m) || is.complex(m)))
+        stop("m must be a numeric or complex matrix")
+    if (!is.matrix(m))
+        m <- as.matrix(m)
+    
+    msvd <- svd(m)
+    if (is.complex(m))
+        msvd$u <- Conj(msvd$u)
+    
+    if( missing(tol) )
+        tol <- max(dim(m))*msvd$d[1]*.Machine$double.eps
+     
+    Positive <- msvd$d > tol  # use only singular values larger than tol
+    if (all(Positive))
+        return( 
+            msvd$v %*% (1/msvd$d * t(msvd$u))
+            )
+    else if (!any(Positive)) 
+        return(
+	    array(0, dim(m)[2:1])
+	    )
+    else 
+        return(
+            msvd$v[, Positive, drop = FALSE] %*%
+	    ((1/msvd$d[Positive]) * t(msvd$u[, Positive, drop = FALSE]))
+	    )
 }
